////////////////////////////////////////////////////////////////
// JFormContainer.java
//  Copyright (C) Neuron Computing 2004-2005. All rights reserved.

package neuron;

import java.awt.*;
import java.awt.event.*;
import java.util.Vector;
import java.lang.System;

////////////////////////////////////////////////////////////////
// JStatic
class JStatic extends Component
{
	protected String m_label;
	protected int m_alignment;

public JStatic(String label, int alignment)
{
	m_label = label;
	m_alignment = alignment;
	setEnabled(false);
}

public void addNotify()
{
}

public void update(Graphics g)
{
	paint(g);
}

public void paint(Graphics g)
{
	FontMetrics fontMetrics = g.getFontMetrics();
	int height = fontMetrics.getHeight()/2;
	int width = fontMetrics.charWidth('M');
	Dimension dim = getSize();
	Color background = g.getColor();
	g.setColor(Color.white);
	g.drawRect(1, height+1, dim.width-1, dim.height-height-1);
	g.setColor(Color.gray);
	g.drawRect(0, height, dim.width-1, dim.height-height-1);
	g.setColor(background);
	g.fillRect(width-1, 0, fontMetrics.stringWidth(m_label)+2, height*2);
	g.setColor(Color.black);
	g.drawString(m_label, width, fontMetrics.getAscent());
	g.setColor(background);
}
};

////////////////////////////////////////////////////////////////
// ControlInfo
class ControlInfo
{
	public boolean m_tabStop;
	public boolean m_group;

public ControlInfo(boolean tabStop, boolean group)
{
	set(tabStop, group);
}

public void set(boolean tabStop, boolean group)
{
	m_tabStop = tabStop;
	m_group = group;
}
};

////////////////////////////////////////////////////////////////
// JFormContainer
public class JFormContainer extends Container
					implements ActionListener, FocusListener
{
	protected String m_title;
	protected Vector m_border = new Vector();
	protected Vector m_info = new Vector();
	protected Vector m_special = new Vector();
	// Cache for the current base units
	protected Dimension m_baseUnits;
	protected Dimension m_fixedSize;

	protected Component m_focusComponent = null;
	protected Button m_defaultButton = null;
	protected String m_result;
	protected Dimension m_margins = new Dimension(8, 8);

	static String s_strCancel = "1";
	static String s_strOK = "2";

public JFormContainer()
{
	m_actionListener = null;
	m_result = s_strCancel;
}
	// Base units
	// x = multiple of 1/4 of the average character width
	//		for the current font
	// y = multiple of 1/8 of the current font height
	static char s_szAlphabet[] = null;

protected Dimension getBaseUnits()
{
	if (m_baseUnits != null)
		return m_baseUnits;
	if (s_szAlphabet == null)
	{
		String alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
		s_szAlphabet = new char[alphabet.length()];
		alphabet.getChars(0, alphabet.length(), s_szAlphabet, 0);
	}
	Graphics g = getGraphics();
	if (g == null)
		return null;

	Dimension dim = new Dimension(0, 0);
	g.setFont(getFont());
	FontMetrics fm = g.getFontMetrics();

	dim.width = fm.charsWidth(s_szAlphabet, 0, 52);
	dim.height = fm.getHeight();
	dim.width = (dim.width + 26)/52;
	dim.width += 1-1;
	m_baseUnits = dim;
	return dim;
}

protected Rectangle convertToPixels(Rectangle rect)
{
	Dimension baseUnits = getBaseUnits();
	Rectangle rectPixels = new Rectangle(0, 0);
	rectPixels.x = (int)(rect.x * (long)baseUnits.width / 4);
	rectPixels.y = (int)(rect.y * (long)baseUnits.height / 8);
	rectPixels.width = (int)(rect.width * (long)baseUnits.width / 4);
	rectPixels.height = (int)(rect.height * (long)baseUnits.height / 8);
	return rectPixels;
}

protected Point convertToPixels(int x, int y)
{
	Dimension baseUnits = getBaseUnits();
	Point pixels = new Point(0, 0);
	pixels.x = (int)(x * (long)baseUnits.width / 4);
	pixels.y = (int)(y * (long)baseUnits.height / 8);
	return pixels;
}

protected Dimension convertToPixels(Dimension dim)
{
	Dimension baseUnits = getBaseUnits();
	Dimension dimOut = new Dimension();
	dimOut.width = (int)(dim.width * (long)baseUnits.width / 4);
	dimOut.height = (int)(dim.height * (long)baseUnits.height / 8);
	return dimOut;
}

protected Rectangle getComponentBounds()
{
	Rectangle rect = null;
	int count = countComponents();
	for (int i=0; i<count; i++)
	{
		Component comp = getComponent(i);
		Rectangle rectComp = comp.getBounds();
		if (rect == null)
			rect = rectComp;
		else
			rect = rectComp.union(rect);
	}
	if (rect == null)
		rect = new Rectangle(0, 0);
	return rect;
}

// Layout attributes
public Dimension getInternalSize()
{
	System.out.println("JFormContainer.getInternalSize");
	Rectangle rect = getComponentBounds();
	Dimension margins = convertToPixels(m_margins);
	return new Dimension(rect.x + rect.width + margins.width,
					rect.y + rect.height + margins.height);
}

public Dimension getFixedSize()
{
	return m_fixedSize;
}

public void setFixedSize(Dimension size)
{
	m_fixedSize = size;
	if (m_fixedSize != null)
	{
		Dimension dim = convertToPixels(m_fixedSize);
		setSize(dim.width, dim.height);
	}
}

public void invalidate()
{
	System.out.println("JFormContainer.invalidate");
	m_baseUnits = null;
	super.invalidate();
}

// Margins in base units
public Dimension getMargins()
{
	return m_margins;
}

public void setMargins(Dimension margins)
{
	m_margins = margins;
}
// Component creation
public int addComponent(Component comp)
{
	super.add(comp);
	ControlInfo info = new ControlInfo(false, false);
	m_info.add(info);
	// Check for buttons
	if (comp instanceof Button)
		((Button)comp).addActionListener(this);
	// Return id
	if (isValid())
		invalidate();
	return m_info.size() - 1;
}

	static final int ok = 2;
	static final int cancel = 1;

public void addOkCancelButtons()
{
	// Ok
	Button button = addButton("Ok", true, JFormContainer.ok);
	setItemProps(0, true, false, false);
	// Cancel
	button = addButton("Cancel", false, JFormContainer.cancel);
	setItemProps(1, true, false, false);
}

public Button addButton(String label, boolean defaultButton, int actionId)
{
	Button button = new Button();
	button.setLabel(label);
	if (defaultButton)
	{
		// There can only be one default push button
		setDefaultButton(button);
	}
	button.setActionCommand(String.valueOf(actionId));
	int id = addComponent(button);
	setItemProps(id, true, false, false);
	return button;
}

public Checkbox addCheckbox(String label)
{
	Checkbox checkbox = new Checkbox();
	checkbox.setLabel(label);
	int id = addComponent(checkbox);
	setItemProps(id, true, false, false);
	return checkbox;
}

public Component addRadioButton(String label)
{
	Checkbox radioButton = new Checkbox();
	// TODO
	radioButton.setCheckboxGroup(null);
	radioButton.setLabel(label);
	int id = addComponent(radioButton);
	setItemProps(id, true, false, false);
	return radioButton;
}

public Label addLabel(String labelText, int alignment)
{
	Label label = new Label();
	label.setText(labelText);
	label.setAlignment(alignment);
	addComponent(label);
	return label;
}

public Component addStatic(String labelText, int alignment)
{
	JStatic staticComp = new JStatic(labelText, alignment);
	addComponent(staticComp);
	m_special.add(staticComp);
	return staticComp;
}

public Scrollbar addScrollbar(int orientation)
{
	Scrollbar sb = new Scrollbar(orientation);
	addComponent(sb);
	return sb;
}

public List addListBox(int rows, boolean multipleMode)
{
	List list = new List(rows);
	list.setBackground(SystemColor.window);
	list.setMultipleMode(multipleMode);
	int id = addComponent(list);
	setItemProps(id, true, false, false);
	return list;
}

public Component addChoice()
{
	Choice choice = new Choice();
	choice.setBackground(SystemColor.window);
	int id = addComponent(choice);
	setItemProps(id, true, false, false);
	return choice;
}

public TextField addTextField(int columns)
{
	TextField text = new TextField(columns);
	text.setBackground(SystemColor.window);
	int id = addComponent(text);
	setItemProps(id, true, false, false);
	return text;
}

public TextArea addTextArea(int rows, int columns)
{
	TextArea text = new TextArea(rows, columns);
	text.setBackground(SystemColor.window);
	int id = addComponent(text);
	setItemProps(id, true, false, false);
	return text;
}

public void setComponentProps(Component comp, boolean tabStop, boolean group, boolean border)
{
	int id = findComponentId(comp);
	setItemProps(id, tabStop, group, border);
}

public void setItemProps(int id, boolean tabStop, boolean group, boolean border)
{
	Component comp = getComponent(id);
	if (comp == null)
		throw new IllegalArgumentException();
	ControlInfo info = (ControlInfo)m_info.get(id);
	info.set(tabStop, group);
	if (border && !(comp instanceof TextComponent))
	{
		m_border.add(comp);
	}
}

public void setDefaultButton(Button button)
{
	m_defaultButton = button;
}

// Methods
	// Layout methods
	static final int left = 0;
	static final int below = 0;
	static final int right = 1;
	static final int above = 2;
	static final int leftBelow = 0;
	static final int rightBelow = 1;
	static final int leftAbove = 2;
	static final int rightAbove = 3;

// Bounds, position, distance in base units
public void setComponentBounds(Component comp, Rectangle rect)
{
	// Convert position
	rect = convertToPixels(rect);
	// Set position
	comp.setBounds(rect.x, rect.y, rect.width, rect.height);
}

public void setItemBounds(int id, Rectangle rect)
{
	Component comp = getComponent(id);
	if (comp == null)
		throw new IllegalArgumentException();
	setComponentBounds(comp, rect);
}

public void setDefaultBounds(int id, int x, int y, int align)
{
	// x and y are in dialog units
	Point pt = convertToPixels(x, y);
	// Delay computing this until needed
	Rectangle rectBounds = null;
	if (x < 0)
	{
		// x < 0 measures distance from the current component bounds
		rectBounds = getComponentBounds();
		pt.x = rectBounds.x + rectBounds.width - pt.x;
	}
	if (y < 0)
	{
		// y < 0 measures distance from the current component bounds
		if (rectBounds == null)
			rectBounds = getComponentBounds();
		pt.y = rectBounds.y + rectBounds.height - pt.y;
	}
	setDefaultPixelBounds(id, pt.x, pt.y, align);
}

public void alignRelative(int id, int relativeTo, int alignX, int alignY)
{
	Component comp2 = getComponent(relativeTo);
	if (comp2 == null)
		throw new IllegalArgumentException();
	Rectangle rect2 = comp2.getBounds();
	Point ptAlign = convertToPixels(alignX, alignY);
	int align = 0;
	// x
	int x;
	if (alignX == 0)
		x = rect2.x;
	else if (alignX > 0)
		x = rect2.x + rect2.width + ptAlign.x;
	else
	{
		x = rect2.x + ptAlign.x;
		align |= right;
	}
	// y
	int y;
	if (alignY == 0)
		y = rect2.y;
	else if (alignY > 0)
		y = rect2.y + rect2.height + ptAlign.y;
	else
	{
		y = rect2.y + ptAlign.y;
		align |= above;
	}
	setDefaultPixelBounds(id, x, y, align);
}

public void alignLeft(int id, int relativeTo, int distance)
{
	alignRelative(id, relativeTo, -distance, 0);
}

public void alignRight(int id, int relativeTo, int distance)
{
	alignRelative(id, relativeTo, distance, 0);
}

public void alignBelow(int id, int relativeTo, int distance)
{
	alignRelative(id, relativeTo, 0, distance);
}

public void alignAbove(int id, int relativeTo, int distance)
{
	alignRelative(id, relativeTo, 0, -distance);
}

	// x and y in pixels
public void setDefaultPixelBounds(int id, int x, int y, int align)
{
	Component comp = getComponent(id);
	if (comp == null)
		throw new IllegalArgumentException();
	Dimension dim = comp.getPreferredSize();
	// x and y are in dialog units
	Point ptSpacing = convertToPixels(x, y);
	Dimension margins = null;
	// Left
	if (x == 0)
	{
		margins = convertToPixels(m_margins);
		x = margins.width;
	}
	if ((align & right) != 0)
		x -= dim.width;
	// Top
	if (y == 0)
	{
		if (margins == null)
			margins = convertToPixels(m_margins);
		y = margins.height;
	}
	if ((align & above) != 0)
		y -= dim.height;
	// Set position
	comp.setBounds(x, y, dim.width, dim.height);
}

// Result
public boolean isOk()
{
	return m_result.equals(s_strOK);
}

public boolean isCanceled()
{
	return m_result.equals(s_strCancel);
}

public boolean isResult(String result)
{
	return m_result.equals(result);
}

public String getResult()
{
	return m_result;
}

public void endDialog(String result)
{
	Container parent = getParent();
	if (parent instanceof Window)
		((Window)parent).hide();
	m_result = result;
}

// Attributes
public void setTitle(String title)
{
	m_title = title;
}

public String getTitle()
{
	return m_title;
}

// Data transfer
public void setText(int id, String text)
{
	Component comp = getComponent(id);
	if (comp == null)
		throw new IllegalArgumentException();
	// Text component
	if (comp instanceof TextComponent)
	{
		((TextComponent)comp).setText(text);
		return;
	}
	// Label
	if (comp instanceof Label)
	{
		((Label)comp).setText(text);
		return;
	}
	// Checkbox
	if (comp instanceof Checkbox)
	{
		((Checkbox)comp).setLabel(text);
		return;
	}
	// Button
	if (comp instanceof Button)
	{
		((Button)comp).setLabel(text);
		return;
	}
}

public String getText(int id)
{
	Component comp = getComponent(id);
	if (comp == null)
		throw new IllegalArgumentException();
	// TextComponent
	if (comp instanceof TextComponent)
	{
		return ((TextComponent)comp).getText();
	}
	// Label
	if (comp instanceof Label)
	{
		return ((Label)comp).getText();
	}
	// Checkbox
	if (comp instanceof Checkbox)
	{
		return ((Checkbox)comp).getLabel();
	}
	// Button
	if (comp instanceof Button)
	{
		return ((Button)comp).getLabel();
	}
	return null;
}

public void setBoolean(int id, boolean value)
{
	Component comp = getComponent(id);
	if (comp == null)
		throw new IllegalArgumentException();
	// Checkbox
	if (comp instanceof Checkbox)
	{
		((Checkbox)comp).setState(value);
	}
}

public boolean getBoolean(int id)
{
	Component comp = getComponent(id);
	if (comp == null)
		throw new IllegalArgumentException();
	// Checkbox
	if (comp instanceof Checkbox)
	{
		return ((Checkbox)comp).getState();
	}
	return false;
}

// Focus helper methods
protected int findComponentId(Component comp)
{
	int count = countComponents();
	for (int i=0; i<count; i++)
	{
		Component c = getComponent(i);
		if (c == comp)
			return i;
	}
	return -1;
}

protected int findNextFocus(int index, boolean back, boolean withinGroup)
{
	// Silently fixes: index < 0 or index > countComponents

	// Find the next component to get focus
	int count = countComponents();
	if (back)
	{
		// Check if we are at the beginning of the current group
		if (withinGroup)
		{
			if (index==0 || index>=count)
				return -1;
			ControlInfo info = (ControlInfo)m_info.get(index);
			if (info.m_group)
				return -1;
		}
		// Find the previous one that is tab stop
		for (int i=index-1; i!=index; i--)
		{
			// Wrap around
			if (i < 0)
				i = count-1;
			ControlInfo info = (ControlInfo)m_info.get(i);
			if (info.m_tabStop)
			{
				// Check that the component is enabled
				Component comp = getComponent(i);
				if (comp.isEnabled() && comp.isVisible())
					return i;
			}
			// Check if we are at the beginning of the current group
			if (withinGroup && info.m_group)
				return -1;
		}
	}
	else
	{
		// Find the next one that is tab stop
		for (int i=index+1; i!=index; i++)
		{
			// Wrap around
			if (i >= count)
				i = 0;
			ControlInfo info = (ControlInfo)m_info.get(i);
			// Check if we passed the end of the current
			// group and went into the next group
			if (withinGroup && (info.m_group || i == 0))
				return -1;
			if (info.m_tabStop)
			{
				// Check that the component is enabled
				Component comp = getComponent(i);
				if (comp.isEnabled() && comp.isVisible())
					return i;
			}
		}
	}
	return -1;
}

protected int findNextFocus(boolean back, boolean withinGroup)
{
	int index;
	if (m_focusComponent == null)
	{
		// Nobody has the focus. Find the first one.
		if (m_defaultButton != null)
		{
			return findComponentId(m_defaultButton);
		}
		// Select the first one
		back = false;
		index = -1;
	}
	else
	{
		// Find the index of the component that currently
		// has the focus
		index = findComponentId(m_focusComponent);
		if (index == -1)
			return -1;
	}
	return findNextFocus(index, back, withinGroup);
}

protected void nextFocus(boolean back, boolean withinGroup)
{
	int index = findNextFocus(back, withinGroup);
	if (index == -1)
		return;
	Component comp = getComponent(index);
	comp.requestFocus();
}

protected void nextFocus(int from, boolean back, boolean withinGroup)
{
	int index = findNextFocus(from, back, withinGroup);
	if (index == -1)
		return;
	Component comp = getComponent(index);
	comp.requestFocus();
}

// IComponent events

// Focus events
public void focusGained(FocusEvent e)
{
	if (e.getSource() == this)
	{
		// We got the focus
		if (m_focusComponent != null)
			m_focusComponent.requestFocus();
		else
			nextFocus(false, false);
		return;
	}
	// See if one of our components got the focus
	Component comp = (Component)e.getSource();
	if (comp.getParent() == this)
		m_focusComponent = comp;
}

public void focusLost(FocusEvent e)
{
	if (e.getSource() == this)
	{
		return;
	}
}

// Keyboard events

void postProcessKeyEvent(KeyEvent e)
{
	System.out.println("JFormContainer.postProcessKeyEvent " + e.toString());

	if (e.getID() != KeyEvent.KEY_PRESSED ||
		e.isControlDown())
	{
		return;
	}

	int key = e.getKeyCode();
	switch (key)
	{
		// Up key maps to vertical scroll code
		case KeyEvent.VK_UP:
//		case Event.LEFT:
			nextFocus(true, true);
			break;
		case KeyEvent.VK_DOWN:
//		case Event.RIGHT:
			nextFocus(false, true);
			break;
		case KeyEvent.VK_TAB:
			nextFocus(e.isShiftDown(), false);
			break;
		case KeyEvent.VK_ENTER:
		{
			if (m_defaultButton != null)
				m_defaultButton.requestFocus();
			// Let it go through
		}
		default:
		{
			return;
		}
	}
	e.consume();
}

// Listeners
	protected ActionListener m_actionListener;

public void addActionListener(ActionListener l)
{
	m_actionListener = l;
}

public void removeActionListener(ActionListener l)
{
	m_actionListener = null;
}

// IActionListener
public void actionPerformed(ActionEvent e)
{
	String strCommand = e.getActionCommand();
	int command = Integer.parseInt(strCommand);
	if (command == JFormContainer.ok || command == JFormContainer.cancel)
	{
		endDialog(strCommand);
	}
	else if (m_actionListener != null)
	{
		m_actionListener.actionPerformed(e);
	}
}

// Painting
protected void draw3DRect(Graphics g, int x, int y, int width, int height, Color crTopLeft, Color crBottomRight)
{
	Color c = g.getColor();

	// The input rectangle is the enclosing rectangle
	// Point(x+widht, y+height) is outside of the rectangle
	g.setColor(crTopLeft);
	g.drawLine(x, y, x, y + height - 2);
	g.drawLine(x + 1, y, x + width - 2, y);
	g.setColor(crBottomRight);
	g.drawLine(x, y + height-1, x + width-1, y + height-1);
	g.drawLine(x + width-1, y, x + width-1, y + height - 2);
	g.setColor(c);
}

protected void draw3DBorder(Graphics g, Rectangle rc)
{
	Color clrBtnShadow = SystemColor.controlShadow;
	Color clrBtnHilite = SystemColor.controlHighlight;
	Color clrWindowFrame = SystemColor.windowBorder;
	Color clrBtnFace = SystemColor.control;

	rc.grow(1, 1);
	draw3DRect(g, rc.x, rc.y, rc.width, rc.height, clrWindowFrame, clrBtnFace);
	rc.grow(1, 1);
	draw3DRect(g, rc.x, rc.y, rc.width, rc.height, clrBtnShadow, clrBtnHilite);
}

public void update(Graphics g)
{
	if (isShowing())
	{
		paint(g);
	}
}

public void paint(Graphics g)
{
	// Paint the default button rectangle
	if (m_defaultButton != null && m_defaultButton.isVisible())
	{
		Rectangle rc = m_defaultButton.getBounds();
		g.drawRect(rc.x-1, rc.y-1, rc.width, rc.height);
	}
	// Paint the border for all controls that had border
	int count = m_border.size();
	for (int i=0; i<count; i++)
	{
		Component comp = (Component)m_border.get(i);
		Rectangle rc = comp.getBounds();
		draw3DBorder(g, rc);
	}
	// Special components
	g.setColor(getBackground());
	count = m_special.size();
	for (int i=0; i<count; i++)
	{
		Component comp = (Component)m_special.get(i);
		if (comp.isVisible())
		{
			Point pt = comp.getLocation();
			g.translate(pt.x, pt.y);
			comp.paint(g);
			g.translate(-pt.x, -pt.y);
		}
	}
}
};
